@@ -1,19 +1,19 @@
 # .. Copyright (C) 2012-2020 Bryan A. Jones.
 #
-#   This file is part of the CodeChat system.
+#   This file is part of the CodeChat System.
 #
-#   The CodeChat system is free software: you can redistribute it and/or
+#   The CodeChat System is free software: you can redistribute it and/or
 #   modify it under the terms of the GNU General Public License as
 #   published by the Free Software Foundation, either version 3 of the
 #   License, or (at your option) any later version.
 #
-#   The CodeChat system is distributed in the hope that it will be
+#   The CodeChat System is distributed in the hope that it will be
 #   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 #   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 #   General Public License for more details.
 #
 #   You should have received a copy of the GNU General Public License
-#   along with the CodeChat system.  If not, see
+#   along with the CodeChat System.  If not, see
 #   <http://www.gnu.org/licenses/>.
 #
 # *******************************
@@ -55,23 +55,17 @@ from thrift.protocol import TBinaryProtocol
 
 # Local application imports
 # -------------------------
-from . import renderer, __version__
+from . import __version__
+from . import render_manager
+from .constants import HTTP_PORT, LOCALHOST, THRIFT_PORT, WEBSOCKET_PORT
 from .gen_py.CodeChat_Services import EditorPlugin
 from .gen_py.CodeChat_Services.ttypes import (
     RenderClientReturn,
     CodeChatClientLocation,
 )
 
-
 # Constants
 # =========
-# The port used for an HTTP connection from the CodeChat Client to the CodeChat Server.
-HTTP_PORT = 5000
-# .. _CodeChat service port:
-#
-# The port used for the Thrift connection between text editor/IDE extensions/plugins and the CodeChat Server. All editor/IDE plugins must use this port to access CodeChat services.
-THRIFT_PORT = 9090
-
 UNKNOWN_CLIENT = "Unknown client id {}."
 
 logger = logging.getLogger(__name__)
@@ -82,7 +76,7 @@ logger = logging.getLogger(__name__)
 # This class implements the EditorPlugin service.
 class CodeChatHandler:
     def __init__(self):
-        self.render_manager: renderer.RenderManager
+        self.render_manager: render_manager.RenderManager
 
     # _`get_client`: Return the HTML for a web client.
     def get_client(self, codeChat_client_location: int) -> RenderClientReturn:
@@ -105,7 +99,7 @@ class CodeChatHandler:
             return ret
 
         # Return what's requested.
-        url = "http://127.0.0.1:{}/client?id={}".format(HTTP_PORT, id)
+        url = "http://{}:{}/client?id={}".format(LOCALHOST, HTTP_PORT, id)
         if codeChat_client_location == CodeChatClientLocation.url:
             # Just return the URL.
             ret_str = url
@@ -124,7 +118,7 @@ class CodeChatHandler:
             )
         elif codeChat_client_location == CodeChatClientLocation.browser:
             # Open in an external browser.
-            webbrowser.open(url, 1)
+            webbrowser.open(url, new=1, autoraise=True)
             ret_str = ""
         else:
             ret = RenderClientReturn(
@@ -137,6 +131,11 @@ class CodeChatHandler:
         logger.info("  => {}".format(ret))
         return ret
 
+    # Indicate the server is alive.
+    def ping(self) -> str:
+        logger.info("ping()\n")
+        return ""
+
     # Render the provided text to HTML, then enqueue it for the web view.
     def start_render(self, text: str, path: str, id: int, is_dirty: bool) -> str:
         logger.info(
@@ -149,6 +148,28 @@ class CodeChatHandler:
             logger.info(" => (empty string)")
             return ""
         else:
+            # If the client specified an non-existant ID, create it.
+            if id < 0:
+                if self.render_manager.threadsafe_create_client(id) != id:
+                    ret = f"Duplicate id {id}."
+                    logger.info(f" => {ret}")
+                    return ret
+                else:
+                    # Since we just created this id, start the CodeChat Client in an external browser.
+                    webbrowser.open(
+                        "http://{}:{}/client?id={}".format(LOCALHOST, HTTP_PORT, id),
+                        new=1,
+                        autoraise=True,
+                    )
+                    # Try the render again.
+                    if self.render_manager.threadsafe_start_render(
+                        text, path, id, is_dirty
+                    ):
+                        # Indicate success.
+                        logger.info(" => (empty string)")
+                        return ""
+
+            # Indicate an error.
             ret = UNKNOWN_CLIENT.format(id)
             logger.info(" => {}".format(ret))
             return ret
@@ -170,12 +191,6 @@ class CodeChatHandler:
         logger.info("  => (empty string)")
         return ""
 
-    # Shut down the server.
-    def shutdown_server(self) -> str:
-        logger.info("shutdown_server")
-        shutdown_event.set()
-        return ""
-
 
 # Instantiate this class, which will be used by both servers.
 handler = CodeChatHandler()
@@ -187,7 +202,7 @@ handler = CodeChatHandler()
 # Server for the CodeChat editor plugin
 # -------------------------------------
 def editor_plugin_server() -> None:
-    transport = TSocket.TServerSocket(host="127.0.0.1", port=THRIFT_PORT)
+    transport = TSocket.TServerSocket(host=LOCALHOST, port=THRIFT_PORT)
     tfactory = TTransport.TBufferedTransportFactory()
     pfactory = TBinaryProtocol.TBinaryProtocolFactory()
     processor = EditorPlugin.Processor(handler)
@@ -221,13 +236,11 @@ client_app = Flask(
 # The endpoint to get the HTML for the CodeChat Client.
 @client_app.route("/client")
 def client_html() -> str:
-    return render_template(
-        "CodeChat_client.html", WEBSOCKET_PORT=renderer.WEBSOCKET_PORT
-    )
+    return render_template("CodeChat_client.html", WEBSOCKET_PORT=WEBSOCKET_PORT)
 
 
-# The endpoint for files requested by a specific client, including rendered source files.
-@client_app.route("/client/<int:id>/<path:url_path>")
+# The endpoint for files requested by a specific client, including rendered source files. Note that ``int`` by default is `positive only <https://werkzeug.palletsprojects.com/en/2.0.x/routing/#werkzeug.routing.IntegerConverter>`_.
+@client_app.route("/client/<int(signed=True):id>/<path:url_path>")
 def client_data(id: int, url_path: str) -> Union[str, Response]:
     # See if we rendered this file.
     html = handler.render_manager.threadsafe_get_render_results(id, url_path)
@@ -268,17 +281,18 @@ def excepthook(type, value, traceback):
 
 # Run both servers. This does not (usually) return.
 def run_servers() -> int:
+
     print(f"The CodeChat Server, v.{__version__}\n")
     logging.basicConfig(level=logging.INFO)
 
     # See if the required ports are in use, probably by another instance of this server.
     if (
         is_port_in_use(HTTP_PORT)
-        or is_port_in_use(renderer.WEBSOCKET_PORT)
+        or is_port_in_use(WEBSOCKET_PORT)
         or is_port_in_use(THRIFT_PORT)
     ):
         print(
-            f"Error: ports {HTTP_PORT}, {renderer.WEBSOCKET_PORT}, and/or {THRIFT_PORT} are already in use.\n"
+            f"Error: ports {HTTP_PORT}, {WEBSOCKET_PORT}, and/or {THRIFT_PORT} are already in use.\n"
             "Hopefully, this means that the CodeChat Server is already running in another process.\n"
             "Exiting.\n"
         )
@@ -287,7 +301,7 @@ def run_servers() -> int:
     # Shut down if any unhandled exception occurs.
     sys.excepthook = excepthook
 
-    # Both servers block when run, so place them in a thread. Mark the Thrift server as a daemon, so it will be killed when the program shuts down.
+    # Both servers block when run, so place them in a thread. Mark the servers as a daemon, so they will be killed when the program shuts down.
     editor_plugin_thread = threading.Thread(
         target=editor_plugin_server, name="Editor plugin server", daemon=True
     )
@@ -300,9 +314,10 @@ def run_servers() -> int:
     flask_server_thread.start()
 
     # Start the render loop in another thread.
-    render_manager = renderer.RenderManager()
-    handler.render_manager = render_manager
-    render_manager_thread = threading.Thread(target=render_manager.run, name="asyncio")
+    handler.render_manager = render_manager.RenderManager(shutdown_event)
+    render_manager_thread = threading.Thread(
+        target=handler.render_manager.run, name="asyncio"
+    )
     render_manager_thread.start()
 
     # On a signal, shut down gracefully.
@@ -317,9 +332,9 @@ def run_servers() -> int:
             shutdown_event.clear()
             break
 
-    print("Shutting down...")
+    logger.info("Shutting down...")
     # This will prevent future editor or web requests from being serviced.
-    render_manager.threadsafe_shutdown()
+    handler.render_manager.threadsafe_shutdown()
     # When this is done, the Flask server and editor plugin sever can be shut down, since they're idle. Since they're daemons, they'll be shut down by exiting main.
     render_manager_thread.join()
     return 0
@@ -329,7 +344,7 @@ def run_servers() -> int:
 def is_port_in_use(port):
     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
         try:
-            s.bind(("localhost", port))
+            s.bind((LOCALHOST, port))
         except OSError:
             return True
         return False
